// SPDX-License-Identifier: GPL-2.0
/*
 * TRIO driver for Mellanox BlueField SoC
 *
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 */

#include <linux/acpi.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/irqdomain.h>
#include <linux/io.h>
#include <linux/kernel.h>
#include <linux/kmod.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/platform_device.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/sysfs.h>

#include "trio_regs.h"

#define DRIVER_NAME		"mlx-trio"
#define DRIVER_VERSION		"0.1"
#define DRIVER_DESCRIPTION	"Mellanox TRIO PCIe host controller driver"

#define TRIO_NUM_IRQS 17
#define L3_PROFILE_NUM	(L3C_PROF_RD_MISS__LENGTH / L3C_PROF_RD_MISS__STRIDE)

/* The PUSH_DMA_EVT_CTR wrapped. */
#define TRIO_PUSH_DMA_EVT_CTR_INT_BIT 10

/* The MAP_EVT_CTR wrapped. */
#define TRIO_MAP_EVT_CTR_INT_BIT 11

struct event_context {
	int event_num;
	int irq;
	struct trio_context *trio;
};

struct trio_context {
	/* The kernel structure representing the device. */
	struct platform_device	*pdev;

	/* Argument to be passed back to the IRQ handler */
	struct event_context *events;

	/* The VA at which the TRIO MMIO registers are mapped */
	char *mmio_base;

	int trio_index;

	/* Name of the bus this TRIO corresponds to */
	const char *bus;

	/* The PCI device this TRIO corresponds to */
	struct pci_dev *trio_pci;

	/* Number of platform_irqs for this device */
	uint32_t num_irqs;
};


static irqreturn_t trio_irq_handler(int irq, void *arg)
{
	struct event_context *ctx = (struct event_context *)arg;
	struct trio_context *trio = ctx->trio;


	pr_err("%s: got IRQ %d event %d on controller %d\n", __func__,
		 irq, ctx->event_num, trio->trio_index);

	return IRQ_HANDLED;
}

static ssize_t l3_cache_profile_show(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	int profile_num;
	struct trio_context *trio;
	struct platform_device *pdev;

	TRIO_DEV_CTL_t tdc;

	pdev = to_platform_device(dev);
	trio = platform_get_drvdata(pdev);

	tdc.word = readq(trio->mmio_base + TRIO_DEV_CTL);

	if (tdc.l3_profile_ovd == 0)
		profile_num = -1;
	else
		profile_num = tdc.l3_profile_val;

	return sprintf(buf, "%d\n", profile_num);
}

static ssize_t l3_cache_profile_store(struct device *dev,
				     struct device_attribute *attr,
				     const char *buf, size_t count)
{
	int err;
	long profile_num;
	struct trio_context *trio;
	struct platform_device *pdev;

	TRIO_DEV_CTL_t tdc;

	pdev = container_of(dev, struct platform_device, dev);
	trio = platform_get_drvdata(pdev);

	err = kstrtol(buf, 10, &profile_num);
	if (err)
		return err;

	if (profile_num == -1) {
		dev_info(dev, "Unlink %s profile\n", trio->bus);

		tdc.word = readq(trio->mmio_base + TRIO_DEV_CTL);
		tdc.l3_profile_ovd = 0;
		writeq(tdc.word, trio->mmio_base + TRIO_DEV_CTL);

	} else if (profile_num < L3_PROFILE_NUM && profile_num >= 0) {
		dev_info(dev, "Change %s to profile %ld\n", trio->bus,
			 profile_num);

		tdc.word = readq(trio->mmio_base + TRIO_DEV_CTL);
		tdc.l3_profile_ovd = 1;
		tdc.l3_profile_val = profile_num;
		writeq(tdc.word, trio->mmio_base + TRIO_DEV_CTL);
	} else
		return -EINVAL;

	return count;
}

static DEVICE_ATTR_RW(l3_cache_profile);

static int trio_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	struct trio_context *trio;
	int i, j, ret, irq;
	int trio_bus, trio_device, trio_function;
	struct resource *res;
	TRIO_INT_SETUP_t int_setup;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res) {
		dev_warn(dev, "%s: failed to find reg resource 0\n", __func__);
		return -ENODEV;
	}

	trio = kzalloc(sizeof(struct trio_context), GFP_KERNEL);
	if (!trio)
		return -ENOMEM;

	platform_set_drvdata(pdev, trio);
	trio->pdev = pdev;
	trio->trio_index = dev->id;

	if (device_property_read_string(dev, "bus_number", &trio->bus)) {
		dev_warn(dev, "%s: failed to retrieve Trio bus name\n",
			 __func__);
		ret = -ENODEV;
		goto err;
	}

	if (device_property_read_u32(dev, "num_irqs", &trio->num_irqs))
		trio->num_irqs = TRIO_NUM_IRQS;
	trio->events = kzalloc(sizeof(struct event_context) * trio->num_irqs,
			       GFP_KERNEL);
	if (!trio->events) {
		ret = -ENOMEM;
		goto err;
	}

	trio->mmio_base = ioremap(res->start, resource_size(res));
	if (IS_ERR(trio->mmio_base)) {
		dev_warn(dev, "%s: ioremap failed for mmio_base %llx err %p\n",
			 __func__, res->start, trio->mmio_base);
		ret = PTR_ERR(trio->mmio_base);
		goto err;
	}

	for (i = 0; i < trio->num_irqs; ++i) {
		struct event_context *ctx = &trio->events[i];
		int dri_ret;

                switch (i) {
                case TRIO_PUSH_DMA_EVT_CTR_INT_BIT:
                case TRIO_MAP_EVT_CTR_INT_BIT:
			/*
			 * These events are not errors, they just indicate
			 * that a performance counter wrapped.  We may want
			 * the performance counter driver to register for them.
			 */
			continue;
                default:
			break;
                }

		irq = platform_get_irq(pdev, i);
		if (irq < 0) {
			dev_warn(dev, "%s: failed to get plat irq %d ret %d\n",
				 __func__, i, irq);
			for (j = i - 1; j >= 0; j--) {
				ctx = &trio->events[j];
				devm_free_irq(&pdev->dev, ctx->irq, ctx);
			}
			ret = -ENXIO;
			goto err;
		}
		ctx->event_num = i;
		ctx->trio = trio;
		ctx->irq = irq;
		dri_ret = devm_request_irq(&pdev->dev, irq, trio_irq_handler, 0,
					   dev_name(dev), ctx);

		dev_dbg(dev, "%s: request_irq returns %d %d->%d\n", __func__,
			 dri_ret, i, irq);
	}

	/* Enable interrupts on this block */
	int_setup.word = 0;
	int_setup.gbl_ena = 1;
	writeq(int_setup.word, trio->mmio_base + TRIO_INT_SETUP);

	/* Create the L3 cache profile on this device */
	device_create_file(dev, &dev_attr_l3_cache_profile);

	/*
	 * Get the corresponding PCI device this trio maps to.
	 * If the bus number can't be read properly, no symlinks are created.
	 */
	if (sscanf(trio->bus, "%d:%d.%d", &trio_bus, &trio_device,
		   &trio_function) != 3) {
		dev_warn(dev, "Device [%s] not valid\n", trio->bus);
		return 0;
	}

	/* The PCI domain/segment would always be 0 here. */
	trio->trio_pci =
		pci_get_domain_bus_and_slot(0, trio_bus,
					    (trio_device << 3) + trio_function);

	/* Add the symlink from the TRIO to the PCI device */
	if (trio->trio_pci != NULL) {
		if (sysfs_create_link(&dev->kobj, &trio->trio_pci->dev.kobj,
				      "pcie_slot")) {
			pci_dev_put(trio->trio_pci);
			trio->trio_pci = NULL;
			dev_warn(dev, "Failed to create symblink for %s\n",
				 trio->bus);
		}
	} else
		dev_warn(dev, "Device %s not found\n", trio->bus);

	return 0;
err:
	dev_warn(dev, "Error probing trio\n");
	if (trio->mmio_base)
		iounmap(trio->mmio_base);
	release_mem_region(res->start, resource_size(res));
	if (trio->events)
		kfree(trio->events);
	kfree(trio);
	platform_set_drvdata(pdev, NULL);
	return ret;
}

static int trio_remove(struct platform_device *pdev)
{
	struct trio_context *trio = platform_get_drvdata(pdev);
	struct device *dev = &pdev->dev;
	struct resource *res;
	int i;

	for (i = 0; i < trio->num_irqs; ++i) {
		struct event_context *ctx = &trio->events[i];

		if (ctx->irq)
			devm_free_irq(&pdev->dev, ctx->irq, ctx);
	}
	device_remove_file(dev, &dev_attr_l3_cache_profile);

	if (trio->mmio_base)
		iounmap(trio->mmio_base);

	/* Delete the symlink and decrement the reference count. */
	if (trio->trio_pci != NULL) {
		sysfs_remove_link(&dev->kobj, "pcie_slot");
		pci_dev_put(trio->trio_pci);
	}
	platform_set_drvdata(pdev, NULL);
	kfree(trio->events);
	kfree(trio);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (res)
		release_mem_region(res->start, resource_size(res));

	return 0;
}

static const struct acpi_device_id trio_acpi_ids[] = {
	{"MLNXBF06", 0},
	{},
};

MODULE_DEVICE_TABLE(acpi, trio_acpi_ids);
static struct platform_driver mlx_trio_driver = {
	.driver = {
		.name = DRIVER_NAME,
		.acpi_match_table = ACPI_PTR(trio_acpi_ids),
	},
	.probe = trio_probe,
	.remove = trio_remove,
};

static int __init trio_init(void)
{
	int ret;

	/*
	 * Request to load the L3 cache driver so that the cache profiles
	 * have been set up by the time the user specify which profile
	 * to use.
	 */
	request_module_nowait("mlx-l3cache");

	ret = platform_driver_register(&mlx_trio_driver);
	if (ret)
		pr_err("Failed to register trio driver.\n");

	return ret;
}

static void __exit trio_exit(void)
{
	platform_driver_unregister(&mlx_trio_driver);
}

module_init(trio_init);
module_exit(trio_exit);

MODULE_DESCRIPTION(DRIVER_DESCRIPTION);
MODULE_AUTHOR("Mellanox Technologies");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);
